@@ -5,9 +5,12 @@
  */
 
 #include "datamanager.h"
-#include "datamanagerlogging.h"
+
 #include "audiomanager.h"
 #include "database.h"
+#include "datamanagerlogging.h"
+#include "entry.h"
+#include "feed.h"
 #include "fetcher.h"
 #include "settingsmanager.h"
 #include <QDateTime>
@@ -145,7 +148,8 @@ Entry *DataManager::getEntry(const QString &id) const
 Entry *DataManager::getEntry(const EpisodeModel::Type type, const int entry_index) const
 {
     QSqlQuery entryQuery;
-    if (type == EpisodeModel::All || type == EpisodeModel::New || type == EpisodeModel::Unread || type == EpisodeModel::Downloaded) {
+    if (type == EpisodeModel::All || type == EpisodeModel::New || type == EpisodeModel::Unread || type == EpisodeModel::Downloading
+        || type == EpisodeModel::PartiallyDownloaded || type == EpisodeModel::Downloaded) {
         if (type == EpisodeModel::New) {
             entryQuery.prepare(QStringLiteral("SELECT id FROM Entries WHERE new=:new ORDER BY updated DESC LIMIT 1 OFFSET :index;"));
             entryQuery.bindValue(QStringLiteral(":new"), true);
@@ -154,11 +158,24 @@ Entry *DataManager::getEntry(const EpisodeModel::Type type, const int entry_inde
             entryQuery.bindValue(QStringLiteral(":read"), false);
         } else if (type == EpisodeModel::All) {
             entryQuery.prepare(QStringLiteral("SELECT id FROM Entries ORDER BY updated DESC LIMIT 1 OFFSET :index;"));
-        } else { // i.e. EpisodeModel::Downloaded
+        } else if (type == EpisodeModel::Downloading) {
             entryQuery.prepare(
                 QStringLiteral("SELECT * FROM Enclosures INNER JOIN Entries ON Enclosures.id = Entries.id WHERE downloaded=:downloaded ORDER BY updated DESC "
                                "LIMIT 1 OFFSET :index;"));
-            entryQuery.bindValue(QStringLiteral(":downloaded"), true);
+            entryQuery.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::Downloading));
+        } else if (type == EpisodeModel::PartiallyDownloaded) {
+            entryQuery.prepare(
+                QStringLiteral("SELECT * FROM Enclosures INNER JOIN Entries ON Enclosures.id = Entries.id WHERE downloaded=:downloaded ORDER BY updated DESC "
+                               "LIMIT 1 OFFSET :index;"));
+            entryQuery.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::PartiallyDownloaded));
+        } else if (type == EpisodeModel::Downloaded) {
+            entryQuery.prepare(
+                QStringLiteral("SELECT * FROM Enclosures INNER JOIN Entries ON Enclosures.id = Entries.id WHERE downloaded=:downloaded ORDER BY updated DESC "
+                               "LIMIT 1 OFFSET :index;"));
+            entryQuery.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::Downloaded));
+        } else {
+            // this should not happen
+            qWarning() << "Cannot find entry type" << type << "in getEntry for entry index" << entry_index;
         }
         entryQuery.bindValue(QStringLiteral(":index"), entry_index);
         Database::instance().execute(entryQuery);
@@ -169,6 +186,7 @@ Entry *DataManager::getEntry(const EpisodeModel::Type type, const int entry_inde
         QString id = entryQuery.value(QStringLiteral("id")).toString();
         return getEntry(id);
     }
+    qWarning() << "Cannot find entry type" << type << "in getEntry for entry index" << entry_index;
     return nullptr;
 }
 
@@ -190,7 +208,8 @@ int DataManager::entryCount(const Feed *feed) const
 int DataManager::entryCount(const EpisodeModel::Type type) const
 {
     QSqlQuery query;
-    if (type == EpisodeModel::All || type == EpisodeModel::New || type == EpisodeModel::Unread || type == EpisodeModel::Downloaded) {
+    if (type == EpisodeModel::All || type == EpisodeModel::New || type == EpisodeModel::Unread || type == EpisodeModel::Downloading
+        || type == EpisodeModel::PartiallyDownloaded || type == EpisodeModel::Downloaded) {
         if (type == EpisodeModel::New) {
             query.prepare(QStringLiteral("SELECT COUNT (id) FROM Entries WHERE new=:new;"));
             query.bindValue(QStringLiteral(":new"), true);
@@ -199,15 +218,25 @@ int DataManager::entryCount(const EpisodeModel::Type type) const
             query.bindValue(QStringLiteral(":read"), false);
         } else if (type == EpisodeModel::All) {
             query.prepare(QStringLiteral("SELECT COUNT (id) FROM Entries;"));
-        } else { // i.e. EpisodeModel::Downloaded
+        } else if (type == EpisodeModel::Downloading) {
             query.prepare(QStringLiteral("SELECT COUNT (id) FROM Enclosures WHERE downloaded=:downloaded;"));
-            query.bindValue(QStringLiteral(":downloaded"), true);
+            query.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::Downloading));
+        } else if (type == EpisodeModel::PartiallyDownloaded) {
+            query.prepare(QStringLiteral("SELECT COUNT (id) FROM Enclosures WHERE downloaded=:downloaded;"));
+            query.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::PartiallyDownloaded));
+        } else if (type == EpisodeModel::Downloaded) {
+            query.prepare(QStringLiteral("SELECT COUNT (id) FROM Enclosures WHERE downloaded=:downloaded;"));
+            query.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::Downloaded));
+        } else {
+            // this should not happen
+            qWarning() << "Cannot find entry type" << type << "in entryCount";
         }
         Database::instance().execute(query);
         if (!query.next())
             return -1;
         return query.value(0).toInt();
     }
+    qWarning() << "Cannot find entry type" << type << "in entryCount";
     return -1;
 }
 
@@ -439,11 +468,15 @@ void DataManager::removeQueueItem(const int index)
     qCDebug(kastsDataManager) << "Queuemap is now:" << m_queuemap;
     // Unset "new" state
     getEntry(m_queuemap[index])->setNew(false);
-    // TODO: Make sure to unset the pointer in the Audio class once it's been
-    // ported to c++
+
+    const QString id = m_queuemap[index];
+
+    // Unload track from AudioManager if it's currently playing
+    if (AudioManager::instance().entry() == getEntry(id)) {
+        AudioManager::instance().setEntry(nullptr);
+    }
 
     // Remove the item from the internal data structure
-    const QString id = m_queuemap[index];
     m_queuemap.removeAt(index);
 
     // Then make sure that the database Queue table reflects these changes
@@ -493,6 +526,24 @@ void DataManager::setLastPlayingEntry(const QString &id)
     Database::instance().execute(query);
 }
 
+void DataManager::deletePlayedEnclosures()
+{
+    QSqlQuery query;
+    query.prepare(QStringLiteral("SELECT * FROM Enclosures INNER JOIN Entries ON Enclosures.id = Entries.id WHERE downloaded=:downloaded AND read=:read;"));
+    query.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::Downloaded));
+    query.bindValue(QStringLiteral(":read"), true);
+    Database::instance().execute(query);
+    while (query.next()) {
+        QString feed = query.value(QStringLiteral("feed")).toString();
+        QString id = query.value(QStringLiteral("id")).toString();
+        qCDebug(kastsDataManager) << "Found entry which has been downloaded and is marked as played; deleting now:" << id;
+        Entry *entry = getEntry(id);
+        if (entry->hasEnclosure()) {
+            entry->enclosure()->deleteFile();
+        }
+    }
+}
+
 void DataManager::importFeeds(const QString &path)
 {
     QUrl url(path);
@@ -575,3 +626,8 @@ void DataManager::updateQueueListnrs() const
         Database::instance().execute(query);
     }
 }
+
+bool DataManager::isFeedExists(const QString &url)
+{
+    return m_feeds.contains(url);
+}
