@@ -5,6 +5,7 @@
  * SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only OR LicenseRef-KDE-Accepted-GPL
  */
 
+#include <KLocalizedString>
 #include <QCryptographicHash>
 #include <QDateTime>
 #include <QDebug>
@@ -21,8 +22,10 @@
 #include <Syndication/Syndication>
 
 #include "database.h"
+#include "enclosure.h"
 #include "fetcher.h"
 #include "fetcherlogging.h"
+#include "kasts-version.h"
 #include "settingsmanager.h"
 
 Fetcher::Fetcher()
@@ -35,6 +38,13 @@ Fetcher::Fetcher()
     manager->setRedirectPolicy(QNetworkRequest::NoLessSafeRedirectPolicy);
     manager->setStrictTransportSecurityEnabled(true);
     manager->enableStrictTransportSecurityStore(true);
+
+#if !defined Q_OS_ANDROID && !defined Q_OS_WIN
+    m_nmInterface = new OrgFreedesktopNetworkManagerInterface(QStringLiteral("org.freedesktop.NetworkManager"),
+                                                              QStringLiteral("/org/freedesktop/NetworkManager"),
+                                                              QDBusConnection::systemBus(),
+                                                              this);
+#endif
 }
 
 void Fetcher::fetch(const QString &url)
@@ -79,6 +89,13 @@ void Fetcher::fetchAll()
 
 void Fetcher::retrieveFeed(const QString &url)
 {
+    if (isMeteredConnection() && !SettingsManager::self()->allowMeteredFeedUpdates()) {
+        Q_EMIT error(Error::Type::MeteredNotAllowed, url, QString(), 0, i18n("Podcast updates not allowed due to user setting"), QString());
+        m_updateProgress++;
+        Q_EMIT updateProgressChanged(m_updateProgress);
+        return;
+    }
+
     qCDebug(kastsFetcher) << "Starting to fetch" << url;
 
     Q_EMIT startedFetchingFeed(url);
@@ -90,7 +107,7 @@ void Fetcher::retrieveFeed(const QString &url)
         if (reply->error()) {
             qWarning() << "Error fetching feed";
             qWarning() << reply->errorString();
-            Q_EMIT error(url, QString(), reply->error(), reply->errorString());
+            Q_EMIT error(Error::Type::FeedUpdate, url, QString(), reply->error(), reply->errorString(), QString());
         } else {
             QByteArray data = reply->readAll();
             Syndication::DocumentSource *document = new Syndication::DocumentSource(data, url);
@@ -339,38 +356,94 @@ void Fetcher::processEnclosure(Syndication::EnclosurePtr enclosure, Syndication:
     query.bindValue(QStringLiteral(":type"), enclosure->type());
     query.bindValue(QStringLiteral(":url"), enclosure->url());
     query.bindValue(QStringLiteral(":playposition"), 0);
-    query.bindValue(QStringLiteral(":downloaded"), false);
+    query.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::Downloadable));
     Database::instance().execute(query);
 }
 
 QString Fetcher::image(const QString &url) const
 {
-    QString path = imagePath(url);
-    if (QFileInfo::exists(path)) {
-        if (QFileInfo(path).size() != 0)
-            return path;
+    if (url.isEmpty()) {
+        return QLatin1String("no-image");
     }
 
-    download(url, path);
+    // if image is already cached, then return the path
+    QString path = imagePath(url);
+    if (QFileInfo::exists(path)) {
+        if (QFileInfo(path).size() != 0) {
+            return QStringLiteral("file://") + path;
+        }
+    }
 
-    return QLatin1String("");
+    // if image has not yet been cached, then check for network connectivity if
+    // possible; and download the image
+    if (canCheckNetworkStatus()) {
+        if (networkConnected() && (!isMeteredConnection() || SettingsManager::self()->allowMeteredImageDownloads())) {
+            download(url, path);
+        } else {
+            return QLatin1String("no-image");
+        }
+    } else {
+        download(url, path);
+    }
+
+    return QLatin1String("fetching");
 }
 
 QNetworkReply *Fetcher::download(const QString &url, const QString &filePath) const
 {
     QNetworkRequest request((QUrl(url)));
     request.setTransferTimeout();
+
+    QFile *file = new QFile(filePath);
+    int resumedAt = 0;
+    if (file->exists() && file->size() > 0) {
+        // try to resume download
+        resumedAt = file->size();
+        qCDebug(kastsFetcher) << "Resuming download at" << resumedAt << "bytes";
+        QByteArray rangeHeaderValue = QByteArray("bytes=") + QByteArray::number(resumedAt) + QByteArray("-");
+        request.setRawHeader(QByteArray("Range"), rangeHeaderValue);
+        file->open(QIODevice::WriteOnly | QIODevice::Append);
+    } else {
+        qCDebug(kastsFetcher) << "Starting new download";
+        file->open(QIODevice::WriteOnly);
+    }
+
+    /*
+    QNetworkReply *headerReply = head(request);
+    connect(headerReply, &QNetworkReply::finished, this, [=]() {
+        if (headerReply->isOpen()) {
+            int fileSize = headerReply->header(QNetworkRequest::ContentLengthHeader).toInt();
+            qCDebug(kastsFetcher) << "Reported download size" << fileSize;
+        }
+        headerReply->deleteLater();
+    });
+    */
+
     QNetworkReply *reply = get(request);
-    connect(reply, &QNetworkReply::finished, this, [=]() {
-        if (reply->isOpen()) {
+
+    connect(reply, &QNetworkReply::readyRead, this, [=]() {
+        if (reply->isOpen() && file) {
             QByteArray data = reply->readAll();
-            QFile file(filePath);
-            file.open(QIODevice::WriteOnly);
-            file.write(data);
-            file.close();
+            file->write(data);
+        }
+    });
+
+    connect(reply, &QNetworkReply::finished, this, [=]() {
+        if (reply->isOpen() && file) {
+            QByteArray data = reply->readAll();
+            file->write(data);
+            file->close();
 
             Q_EMIT downloadFinished(url);
         }
+
+        // clean up; close file if still open in case something has gone wrong
+        if (file) {
+            if (file->isOpen()) {
+                file->close();
+            }
+            delete file;
+        }
         reply->deleteLater();
     });
 
@@ -401,6 +474,57 @@ QString Fetcher::enclosurePath(const QString &url) const
 
 QNetworkReply *Fetcher::get(QNetworkRequest &request) const
 {
-    request.setRawHeader("User-Agent", "Kasts/0.1; Syndication");
+    setHeader(request);
     return manager->get(request);
 }
+
+QNetworkReply *Fetcher::head(QNetworkRequest &request) const
+{
+    setHeader(request);
+    return manager->head(request);
+}
+
+void Fetcher::setHeader(QNetworkRequest &request) const
+{
+    request.setRawHeader(QByteArray("User-Agent"), QByteArray("Kasts/") + QByteArray(KASTS_VERSION_STRING) + QByteArray("; Syndication"));
+}
+
+bool Fetcher::canCheckNetworkStatus() const
+{
+#if !defined Q_OS_ANDROID && !defined Q_OS_WIN
+    qCDebug(kastsFetcher) << "Can NetworkManager be reached?" << m_nmInterface->isValid();
+    return (m_nmInterface && m_nmInterface->isValid());
+#else
+    return false;
+#endif
+}
+
+bool Fetcher::networkConnected() const
+{
+#if !defined Q_OS_ANDROID && !defined Q_OS_WIN
+    qCDebug(kastsFetcher) << "Network connected?" << (m_nmInterface->state() >= 70) << m_nmInterface->state();
+    return (m_nmInterface && m_nmInterface->state() >= 70);
+#else
+    return true;
+#endif
+}
+
+bool Fetcher::isMeteredConnection() const
+{
+#if !defined Q_OS_ANDROID && !defined Q_OS_WIN
+    if (canCheckNetworkStatus()) {
+        // Get network connection status through DBus (NetworkManager)
+        // state == 1: explicitly configured as metered
+        // state == 3: connection guessed as metered
+        uint state = m_nmInterface->metered();
+        qCDebug(kastsFetcher) << "Network Status:" << state;
+        qCDebug(kastsFetcher) << "Connection is metered?" << (state == 1 || state == 3);
+        return (state == 1 || state == 3);
+    } else {
+        return false;
+    }
+#else
+    // TODO: get network connection type for Android and windows
+    return false;
+#endif
+}
