@@ -7,12 +7,19 @@
 
 #pragma once
 
+#include <QFile>
 #include <QNetworkAccessManager>
 #include <QNetworkReply>
 #include <QObject>
 #include <QUrl>
 #include <Syndication/Syndication>
 
+#include "error.h"
+
+#if !defined Q_OS_ANDROID && !defined Q_OS_WIN
+#include "NMinterface.h"
+#endif
+
 class Fetcher : public QObject
 {
     Q_OBJECT
@@ -34,9 +41,15 @@ public:
     Q_INVOKABLE QString image(const QString &url) const;
     void removeImage(const QString &url);
     Q_INVOKABLE QNetworkReply *download(const QString &url, const QString &fileName) const;
-    QNetworkReply *get(QNetworkRequest &request) const;
+
     QString imagePath(const QString &url) const;
     QString enclosurePath(const QString &url) const;
+    QNetworkReply *get(QNetworkRequest &request) const;
+
+    // Network status related methods
+    Q_INVOKABLE bool canCheckNetworkStatus() const;
+    bool networkConnected() const;
+    Q_INVOKABLE bool isMeteredConnection() const;
 
 Q_SIGNALS:
     void startedFetchingFeed(const QString &url);
@@ -48,7 +61,7 @@ Q_SIGNALS:
                             const QString &description,
                             const QDateTime &lastUpdated);
     void feedUpdateFinished(const QString &url);
-    void error(const QString &url, const QString &id, const int errorId, const QString &errorString);
+    void error(Error::Type type, const QString &url, const QString &id, const int errorId, const QString &errorString, const QString &title);
     void entryAdded(const QString &feedurl, const QString &id);
     void downloadFinished(QString url) const;
 
@@ -68,8 +81,15 @@ private:
     void processAuthor(const QString &url, const QString &entryId, const QString &authorName, const QString &authorUri, const QString &authorEmail);
     void processEnclosure(Syndication::EnclosurePtr enclosure, Syndication::ItemPtr entry, const QString &feedUrl);
 
+    QNetworkReply *head(QNetworkRequest &request) const;
+    void setHeader(QNetworkRequest &request) const;
+
     QNetworkAccessManager *manager;
     int m_updateProgress;
     int m_updateTotal;
     bool m_updating;
+
+#if !defined Q_OS_ANDROID && !defined Q_OS_WIN
+    OrgFreedesktopNetworkManagerInterface *m_nmInterface;
+#endif
 };
